"use strict";

{
	// In the C3 runtime's worker mode, all the runtime scripts (e.g. plugin.js, instance.js, actions.js)
	// are loaded in a Web Worker, which has no access to the document so cannot make DOM calls. To help
	// plugins use DOM elements the runtime internally manages a postMessage() bridge wrapped in some helper
	// classes designed to manage DOM elements. Then this script (domSide.js) is loaded in the main document
	// (aka the main thread) where it can make any DOM calls on behalf of the runtime. Conceptually the two
	// ends of the messaging bridge are the "Runtime side" in a Web Worker, and the "DOM side" with access
	// to the Document Object Model (DOM). The addon's plugin.js specifies to load this script on the
	// DOM side by making the call: this._info.SetDOMSideScripts(["c3runtime/domSide.js"])
	// Note that when NOT in worker mode, this entire framework is still used identically, just with both
	// the runtime and the DOM side in the main thread. This allows non-worker mode to work the same with
	// no additional code changes necessary. However it's best to imagine that the runtime side is in a
	// Web Worker, since that is when it is necessary to separate DOM calls from the runtime.
	
	// NOTE: use a unique DOM component ID to ensure it doesn't clash with anything else
	// This must also match the ID in instance.js and plugin.js.
	const DOM_COMPONENT_ID = "shepherdgames-rive";

	function StopPropagation(e)
	{
		e.stopPropagation();
	}

	const HANDLER_CLASS = class MyDOMHandler extends globalThis.DOMElementHandler
	{
		constructor(iRuntime)
		{
			super(iRuntime, DOM_COMPONENT_ID);
		}

		CreateElement(elementId, e)
		{
			const elem = document.createElement("canvas");
			elem.style.position = "absolute"

			// Prevent touches reaching the canvas
			elem.addEventListener("touchstart", StopPropagation);
			elem.addEventListener("touchmove", StopPropagation);
			elem.addEventListener("touchend", StopPropagation);
			
			// Prevent clicks being blocked
			elem.addEventListener("mousedown", StopPropagation);
			elem.addEventListener("mouseup", StopPropagation);
			
			// Prevent key presses being blocked by the Keyboard object
			elem.addEventListener("keydown", StopPropagation);
			elem.addEventListener("keyup", StopPropagation);

			// Attach click event which posts a message to the runtime telling it that this event happened.
			elem.addEventListener("click", () => this.PostToRuntimeElement("click", elementId /*, {
				// optional object in third parameter with extra details
			}*/));

			// The create message includes the state retrieved by GetElementState() in instance.js,
			// so also update the element state based on those details.
			this.UpdateState(elem, e);

			return elem;
		}

		UpdateState(elem, e)
		{
			// Update the state of the DOM element 'elem' with the state 'e'. The state has been
			// retrieved by calling GetElementState() in instance.js, which includes all necessary
			// details to set the correct state of the DOM element.
			// NOTE: the runtime automatically manages the position, size and visibility of the DOM
			// element, so this only needs to handle state unique to the element, such as the button
			// text in this case.
			elem.textContent = e["text"];
			const filePath = e["filePath"];
			const stateMachine = e["stateMachine"];

			// console.log(filePath, stateMachine);
			
			if(filePath != "" && stateMachine != ""){
				console.log("Initializing Rive with:", { filePath, stateMachine });
				const rive = globalThis.rive;
				const r = new rive.Rive({
					src: filePath,
					canvas: elem,
					autoplay: true,
					stateMachines: stateMachine,
					onLoad: () => {
						console.log("Rive file loaded successfully");
						r.resizeDrawingSurfaceToCanvas();
					},
					onError: (err) => {
						console.error("Rive error:", err);
					}
				});
				
				const eventType = rive.EventType;
				const riveEventType = rive.RiveEventType;

				function onRiveEventReceived(riveEvent) {
					console.log("Rive event received:", riveEvent);
					const eventData = riveEvent.data;
					const eventProperties = eventData.properties;
					
					// Forward the event to Construct 3's event system
					if (eventData.type === riveEventType.General) {
						console.log("Dispatching general event:", eventData.name);
						// Create a custom event with the Rive event data
						const customEvent = new CustomEvent('riveEvent', {
							detail: {
								name: eventData.name,
								properties: eventProperties
							}
						});
						elem.dispatchEvent(customEvent);
					} else if (eventData.type === riveEventType.OpenUrl) {
						console.log("Dispatching URL event:", eventData.url);
						// Handle URL events
						window.open(eventData.url);
					}
				}

				// Add event listener for Rive events
				r.on(eventType.RiveEvent, onRiveEventReceived);
				console.log("Added Rive event listener");

				// Add event listener for state changes
				r.on(eventType.StateChange, (event) => {
					console.log("State change detected:", event);
					const customEvent = new CustomEvent('riveStateChange', {
						detail: event
					});
					elem.dispatchEvent(customEvent);
				});
				console.log("Added state change listener");

				// Store the Rive instance on the element for later access
				elem.riveInstance = r;
				console.log("Rive instance stored on element");

				// Add click handler to the canvas
				elem.addEventListener('click', (e) => {
					console.log("Canvas clicked");
					// Get the Rive instance
					const riveInstance = elem.riveInstance;
					if (riveInstance) {
						// Get the artboard
						const artboard = riveInstance.artboard;
						if (artboard) {
							// Get the state machine
							const stateMachine = riveInstance.animator.stateMachines[0];
							if (stateMachine) {
								console.log("State machine found:", stateMachine);
								
								// Get canvas bounds
								const rect = elem.getBoundingClientRect();
								
								// Get click coordinates relative to the canvas
								const canvasX = e.clientX - rect.left;
								const canvasY = e.clientY - rect.top;
								
								// Get the forward matrix for coordinate transformation
								const forwardMatrix = riveInstance.runtime.computeAlignment(
									riveInstance._layout.runtimeFit(riveInstance.runtime),
									riveInstance._layout.runtimeAlignment(riveInstance.runtime),
									{
										minX: 0,
										minY: 0,
										maxX: rect.width,
										maxY: rect.height
									},
									artboard.bounds
								);
								
								// Create inverse matrix for coordinate transformation
								const invertedMatrix = new riveInstance.runtime.Mat2D();
								forwardMatrix.invert(invertedMatrix);
								
								// Create vectors for coordinate transformation
								const canvasCoordinatesVector = new riveInstance.runtime.Vec2D(canvasX, canvasY);
								const transformedVector = riveInstance.runtime.mapXY(invertedMatrix, canvasCoordinatesVector);
								
								// Get transformed coordinates
								const transformedX = transformedVector.x();
								const transformedY = transformedVector.y();
								
								console.log("Transformed coordinates:", { transformedX, transformedY });
								
								// Clean up
								transformedVector.delete();
								invertedMatrix.delete();
								canvasCoordinatesVector.delete();
								forwardMatrix.delete();
								
								// Forward the click to the state machine with transformed coordinates
								stateMachine.instance.pointerDown(transformedX, transformedY);
								stateMachine.instance.pointerUp(transformedX, transformedY);
							}
						}
					}
				});
			}
			
		}
	};
	
	globalThis.RuntimeInterface.AddDOMHandlerClass(HANDLER_CLASS);
}